from __future__ import annotations

import os
import warnings

import numpy as np
from packaging.version import Version

import yt
from yt.utilities.exceptions import YTUnidentifiedDataType
from yt_idefix._io import vtk_io
from yt_idefix.data_structures import IdefixVtkDataset, PlutoVtkDataset

__all__ = ["load", "load_stretched"]


YT_VERSION = Version(yt.__version__)


class VisibleDeprecationWarning(Warning):
    pass


class IdefixVtkDatasetSeries(yt.DatasetSeries):
    _dataset_cls = IdefixVtkDataset


def load(fn, *args, **kwargs):
    msg = (
        "yt_idefix.load is a diminished wrapper for to yt.load, "
        "and is not necessary with yt > 4.0.2\nplease "
    )

    if YT_VERSION <= Version("4.0.2"):
        msg += "upgrade yt and "
    msg += "use yt.load instead"
    msg += (
        "yt_idefix.load was deprecated in version 0.11.0 "
        "and will be removed completely in a future version"
    )
    warnings.warn(
        msg,
        category=VisibleDeprecationWarning,
        stacklevel=2,
    )

    fn = os.path.expanduser(fn)
    if not os.path.exists(fn):
        raise FileNotFoundError(fn)

    if YT_VERSION < Version("4.0.2") and fn.endswith(".vtk"):
        if any(wildcard in fn for wildcard in "[]?!*"):
            return IdefixVtkDatasetSeries(fn, *args, **kwargs)
        elif IdefixVtkDataset._is_valid(fn, *args, **kwargs):
            return IdefixVtkDataset(fn, *args, **kwargs)
        elif PlutoVtkDataset._is_valid(fn, *args, **kwargs):
            return PlutoVtkDataset(fn, *args, **kwargs)
        else:
            raise YTUnidentifiedDataType(fn, *args, **kwargs)
    else:
        return yt.load(fn, *args, **kwargs)


def load_stretched(fn, *, geometry: str | None = None, **kwargs):
    """
    A small, specialized wrapper around yt.load_hexahedral_mesh
    Compatible with streched grids but comes at a significant cost as of yt 4.0
    - limited support for common operations such as projections
    - fields are not properly dimensioned
    - no lazy-loading (all data has to reside in memory)
    - only supports vtk outputs (not dumps)
    """

    # brute force validation
    if not (
        IdefixVtkDataset._is_valid(fn, geometry=geometry)
        or PlutoVtkDataset._is_valid(fn, geometry=geometry)
    ):
        raise TypeError(
            "yt_idefix.load_stretched only supports Idefix and Pluto vtk files"
        )

    # actual parsing
    with open(fn, "rb") as fh:
        md = vtk_io.read_metadata(fh)
        coords = vtk_io.read_grid_coordinates(
            fh, geometry=geometry or md.get("geometry")
        ).padded()
        field_offset_index = vtk_io.read_field_offset_index(
            fh, shape=coords.array_shape
        )

    if geometry is None:
        geometry = md["geometry"]

    data: dict[str, np.ndarray] = {}
    with open(fn, "rb") as fh:
        for name, offset in field_offset_index.items():
            data[name] = vtk_io.read_single_field(
                fh, shape=coords.array_shape, offset=offset, skip_data=False
            )

    coordinates, connectivity = yt.hexahedral_connectivity(
        coords.x1, coords.x2, coords.x3
    )

    bbox = np.array(
        [
            [coords.x1[0], coords.x1[-1]],
            [coords.x2[0], coords.x2[-1]],
            [coords.x3[0], coords.x3[-1]],
        ]
    )

    return yt.load_hexahedral_mesh(
        data,
        coordinates=coordinates,
        connectivity=connectivity,
        bbox=bbox,
        geometry=geometry,
        **kwargs,
    )



from typing import Any, Dict

import numpy as np

# Physical constants in c.g.s units defined in $PLUTO_DIR/Src/pluto.h
# The values are copied from PLUTO 4.4-patch-2
pluto_def_constants = {
    "CONST_AH": 1.008,
    "CONST_AHe": 4.004,
    "CONST_AZ": 30.0,
    "CONST_amu": 1.66053886e-24,
    "CONST_au": 1.49597892e13,
    "CONST_c": 2.99792458e10,
    "CONST_e": 4.80320425e-10,
    "CONST_eV": 1.602176463158e-12,
    "CONST_G": 6.6726e-8,
    "CONST_h": 6.62606876e-27,
    "CONST_kB": 1.3806505e-16,
    "CONST_ly": 0.9461e18,
    "CONST_mp": 1.67262171e-24,
    "CONST_mn": 1.67492728e-24,
    "CONST_me": 9.1093826e-28,
    "CONST_mH": 1.6733e-24,
    "CONST_Msun": 2.0e33,
    "CONST_Mearth": 5.9736e27,
    "CONST_NA": 6.0221367e23,
    "CONST_pc": 3.0856775807e18,
    "CONST_PI": 3.14159265358979,
    "CONST_Rearth": 6.378136e8,
    "CONST_Rgas": 8.3144598e7,
    "CONST_Rsun": 6.96e10,
    "CONST_sigma": 5.67051e-5,
    "CONST_sigmaT": 6.6524e-25,
}


class _PlutoBaseUnits:
    """Derive base units from a given combination of units"""

    _base_unit_list = ("mass_unit", "length_unit", "time_unit")

    def __init__(self, unit_combination: Dict[str, Any]):
        # Fow now, unit_combination has been validated before passed in
        # But we still need to check the number of units here for insurance
        if len(unit_combination) != 3:
            raise ValueError(
                "_PlutoBaseUnits requires a combination of 3 units "
                f"({len(unit_combination)} given)"
            )
        self.unit_combination = unit_combination
        self._unit_cache = self.unit_combination
        _data = {}
        for unit in self._base_unit_list:
            self._setup_unit(unit)
            _data[unit] = self._unit_cache[unit]
        self._data = _data

    def __getitem__(self, key):
        return self._data[key]

    # Some unit_setup_functions
    # The condition statements are essential to prevent getting stuck in infinite recursion
    # However they are fragile, one must be cautious to change them
    def _setup_time_unit(self):
        uc = self._unit_cache
        if "time_unit" in uc:
            return
        time = uc["length_unit"] / uc["velocity_unit"]
        uc["time_unit"] = time

    def _setup_length_unit(self):
        uc = self._unit_cache
        if "length_unit" in uc:
            return
        if "time_unit" in uc:
            length = uc["velocity_unit"] * uc["time_unit"]
        else:
            length = (uc["mass_unit"] / uc["density_unit"]) ** (1 / 3)
        uc["length_unit"] = length

    def _setup_mass_unit(self):
        uc = self._unit_cache
        if "mass_unit" in uc:
            return
        if "density_unit" in uc:
            mass = uc["density_unit"] * uc["length_unit"] ** 3
        else:
            mass = (
                (uc["magnetic_unit"] / uc["velocity_unit"]) ** 2
                / 4.0
                / np.pi
                * uc["length_unit"] ** 3
            )
        uc["mass_unit"] = mass

    def _setup_velocity_unit(self):
        uc = self._unit_cache
        if "velocity_unit" in uc:
            return
        if "length_unit" in uc and "time_unit" in uc:
            vel = uc["length_unit"] / uc["time_unit"]
        elif "magnetic_unit" in uc:
            vel = uc["magnetic_unit"] / np.sqrt(4.0 * np.pi * uc["density_unit"])
        else:
            vel = (uc["mass_unit"] / uc["density_unit"]) ** (1 / 3) / uc["time_unit"]
        uc["velocity_unit"] = vel

    def _setup_density_unit(self):
        uc = self._unit_cache
        if "density_unit" in uc:
            return
        if "length_unit" in uc:
            density = uc["mass_unit"] / uc["length_unit"] ** 3
        elif "velocity_unit" in uc:
            density = (uc["magnetic_unit"] / uc["velocity_unit"]) ** 2 / 4.0 / np.pi
        else:
            density = (
                (uc["magnetic_unit"] * uc["time_unit"]) ** 3 / uc["mass_unit"]
            ) ** 2 / (4.0 * np.pi) ** 3
        uc["density_unit"] = density

    def _setup_unit(self, unit):
        """Calculate a unit according to the given unit setup functions"""
        f_name = "_setup_" + unit
        f = getattr(self, f_name)
        try:
            f()
        except KeyError as err:
            # If any unit in the function is missing, calculate that first,
            # then come back and try it again.
            missing_unit = err.args[0]
            self._setup_unit(missing_unit)
            self._setup_unit(unit)



from abc import ABC, abstractmethod
from typing import BinaryIO, Tuple, cast

import numpy as np

from yt.utilities.io_handler import BaseIOHandler

from ._io import dmp_io, vtk_io


class IdefixIOHandler(BaseIOHandler, ABC):
    _particle_reader = False
    _dataset_type = "idefix-NONE"

    def __init__(self, ds):
        BaseIOHandler.__init__(self, ds)
        self.ds = ds
        self._data_file = ds.parameter_filename

    def _read_fluid_selection(self, chunks, selector, fields, size):
        data = {}

        for field in fields:
            data[field] = np.empty(size, dtype="float64")

        with open(self._data_file, mode="rb") as fh:
            ind = 0
            for chunk in chunks:
                for grid in chunk.objs:
                    nd = 0
                    for field in fields:
                        ftype, fname = field
                        foffset = grid._index._field_offsets[fname]
                        values = self._read_single_field(fh, foffset)
                        nd = grid.select(selector, values, data[field], ind)
                    ind += nd
        return data

    def _read_particle_coords(self, chunks, ptf):
        # This needs to *yield* a series of tuples of (ptype, (x, y, z)).
        # chunks is a list of chunks, and ptf is a dict where the keys are
        # ptypes and the values are lists of fields.
        raise NotImplementedError("Particles are not currently supported for Idefix")

    def _read_particle_fields(self, chunks, ptf, selector):
        # idefix doesn't have particles (yet)
        raise NotImplementedError("Particles are not currently supported for Idefix")

    def _read_chunk_data(self, chunk, fields):
        # This reads the data from a single chunk without doing any selection,
        # and is only used for caching data that might be used by multiple
        # different selectors later. For instance, this can speed up ghost zone
        # computation.
        raise NotImplementedError

    # The following methods are frontend-specific
    @abstractmethod
    def _read_single_field(self, fh: BinaryIO, offset: int) -> np.ndarray:
        pass


class IdefixVtkIOHandler(IdefixIOHandler):
    _dataset_type = "idefix-vtk"

    def _read_single_field(self, fh: BinaryIO, offset: int) -> np.ndarray:
        shape = cast(Tuple[int, int, int], tuple(self.ds.domain_dimensions))
        return vtk_io.read_single_field(fh, shape=shape, offset=offset, skip_data=False)


class IdefixDmpIOHandler(IdefixIOHandler):
    _dataset_type = "idefix-dmp"

    def _read_single_field(self, fh: BinaryIO, offset: int) -> np.ndarray:
        return dmp_io.read_single_field(fh, offset)



from __future__ import annotations

import struct
import warnings
from typing import Any, BinaryIO, Literal, overload

import numpy as np

from .commons import ByteSize, Coordinates, Shape

KNOWN_GEOMETRIES = {0: "cartesian", 1: "polar", 2: "spherical", 3: "cylindrical"}


def read_header(filename: str) -> str:
    with open(filename, "rb") as fh:
        return "".join(fh.readline(256).decode() for _ in range(2))


@overload
def read_single_field(
    fh: BinaryIO,
    *,
    shape: tuple[int, int, int],
    offset: int | None = None,
    skip_data: Literal[False],
) -> np.ndarray:
    ...


@overload
def read_single_field(
    fh: BinaryIO,
    *,
    shape: tuple[int, int, int],
    offset: int | None = None,
    skip_data: Literal[True],
) -> None:
    ...


def read_single_field(
    fh,
    *,
    shape,
    offset=None,
    skip_data=False,
):
    count = np.prod(shape)
    if offset is not None and fh.tell() != offset:
        fh.seek(offset)
    if skip_data:
        fh.seek(count * ByteSize.FLOAT, 1)
        data = None
    else:
        data = np.fromfile(fh, ">f", count=count)
        data.shape = shape[::-1]
        data = data.T
    return data


def read_shape(s: str) -> Shape:
    # read a specific line containing nx, ny, nz

    assert s.startswith("DIMENSIONS")
    raw: list[str] = s.split()[1:]
    if len(raw) != 3:
        raise RuntimeError
    return Shape(*(int(_) for _ in raw))


def parse_shape(s: str, md: dict[str, Any]) -> None:
    md["shape"] = read_shape(s)


# this may not be kept in the following form
def read_metadata(fh: BinaryIO) -> dict[str, Any]:

    fh.seek(0)
    # skip over the first 4 lines which normally contains
    # VTK DataFile Version x.x
    # <Comments>
    # BINARY
    # DATASET RECTILINEAR_GRID or STRUCTURED_GRID
    for _ in range(4):
        next(fh)

    metadata: dict[str, Any] = {}
    line = next(fh).decode()  # DIMENSIONS NX NY NZ or FIELD
    if line.startswith("FIELD"):
        # Idefix >= 0.8
        nfield = int(line.split()[2])
        for _ in range(nfield):
            d = next(fh).decode()
            if d.startswith("GEOMETRY"):
                geom_flag: int = struct.unpack(">i", fh.read(4))[0]
                geometry_from_data = KNOWN_GEOMETRIES.get(geom_flag)
                if geometry_from_data is None:
                    warnings.warn(
                        f"Unknown geometry enum value {geom_flag}, please report this."
                    )
                metadata["geometry"] = geometry_from_data
            elif d.startswith("TIME"):
                metadata["time"] = struct.unpack(">f", fh.read(4))[0]
            elif d.startswith("PERIODICITY"):
                metadata["periodicity"] = tuple(
                    np.fromfile(fh, dtype=">i4", count=3).astype(bool)
                )
            else:
                warnings.warn(f"Found unknown field {d!r}")
            next(fh)  # skip extra linefeed (empty line)
        parse_shape(next(fh).decode(), metadata)

    elif line.startswith("DIMENSIONS"):
        parse_shape(line, metadata)

    else:
        raise RuntimeError(f"Failed to parse {line!r}")

    return metadata


def read_grid_coordinates(
    fh: BinaryIO,
    *,
    geometry: str | None = None,
) -> Coordinates:
    # Return cell edges coordinates

    fh.seek(0)
    md = read_metadata(fh)

    geometry = md.get("geometry", geometry)

    shape = md["shape"]
    coords: list[np.ndarray] = []
    # now assuming that fh is positioned at the end of metadata
    if geometry in ("cartesian", "cylindrical"):
        # In Idefix, cylindrical geometry is only meant to be used in 2D,
        # so the grid structure is effectively cartesian (R, z)
        for nx in shape:
            next(fh)
            coords.append(np.fromfile(fh, dtype=">f", count=nx))
            next(fh)
        line = next(fh).decode()
        point_type, npoints = (t(_) for t, _ in zip((str, int), line.split()))
        next(fh)
        if point_type == "CELL_DATA":
            array_shape = shape.to_cell_centered()
        else:
            if point_type == "POINT_DATA":
                # raw data is cell center coords, extrapolation would be needed here
                # However it's probably never going to be worth it since this branch
                # corresponds to Idefix < 0.8 and should never be hit in practice.
                # Raising a warning should suffice
                warnings.warn(
                    "point_type=POINT_DATA case is not fully supported. "
                    "Domain edges will be slightly off."
                )
            else:
                warnings.warn(
                    f"Got unexpected value point_type={point_type!r}. "
                    "Results are not guaranteed."
                )
            array_shape = shape

    elif geometry in ("polar", "spherical"):
        rshape = Shape(*reversed(shape))
        npoints = int(next(fh).decode().split()[1])  # POINTS NXNYNZ float
        assert shape.size == npoints
        points = np.fromfile(fh, dtype=">f", count=3 * npoints)
        next(fh)

        xcart = points[::3]
        xcart.shape = rshape
        xcart = xcart.T

        ycart = points[1::3]
        ycart.shape = rshape
        ycart = ycart.T

        zcart = points[2::3]
        zcart.shape = rshape
        zcart = zcart.T

        # Reconstruct the polar coordinate system
        if geometry == "polar":
            # on disk coordinates are cell face coordinates.
            # We need to convert from cartesian to polar,
            # but no interpolation is needed.
            r = np.sqrt(xcart[:, 0, 0] ** 2 + ycart[:, 0, 0] ** 2)
            theta = np.unwrap(np.arctan2(ycart[0, :, 0], xcart[0, :, 0]))
            z = zcart[0, 0, :]

            data_type = next(fh).decode().split()[0]  # CELL_DATA (NX-1)(NY-1)(NZ-1)
            assert data_type == "CELL_DATA"
            next(fh)

            # manually changing phase origin (theta) to match
            # results from Idefix's pytools
            coords = [r, theta + np.pi, z]
        elif geometry == "spherical":
            # Reconstruct the spherical coordinate system
            if shape.n3 == 1:
                r = np.sqrt(xcart[:, 0, 0] ** 2 + ycart[:, 0, 0] ** 2)
                phi = np.unwrap(
                    np.arctan2(zcart[0, shape.n2 // 2, :], xcart[0, shape.n2 // 2, :])
                )
                theta = np.arccos(
                    ycart[0, :, 0] / np.sqrt(xcart[0, :, 0] ** 2 + ycart[0, :, 0] ** 2)
                )
            else:
                r = np.sqrt(
                    xcart[:, 0, 0] ** 2 + ycart[:, 0, 0] ** 2 + zcart[:, 0, 0] ** 2
                )
                phi = np.unwrap(
                    np.arctan2(
                        ycart[shape.n1 // 2, shape.n2 // 2, :],
                        xcart[shape.n1 // 2, shape.n2 // 2, :],
                    )
                )
                theta = np.arccos(
                    zcart[0, :, 0]
                    / np.sqrt(
                        xcart[0, :, 0] ** 2 + ycart[0, :, 0] ** 2 + zcart[0, :, 0] ** 2
                    )
                )
            coords = [r, theta, phi]

            data_type = next(fh).decode().split()[0]  # CELL_DATA (NX-1)(NY-1)(NZ-1)
            if data_type != "CELL_DATA":
                raise RuntimeError
            next(fh)
        else:
            raise RuntimeError("This should be logically impossible.")

        array_shape = shape.to_cell_centered()

    else:
        raise RuntimeError(f"Found unknown geometry {geometry!r}")

    return Coordinates(coords[0], coords[1], coords[2], array_shape)


def read_field_offset_index(fh: BinaryIO, shape: Shape) -> dict[str, int]:
    # assuming fh is correctly positioned (read_grid_coordinates must be called first)
    retv: dict[str, int] = {}

    while True:
        line = fh.readline()
        if len(line) < 2:
            break
        s = line.decode()
        datatype, varname, dtype = s.split()

        # some versions of Pluto define field names in lower case
        # so we normalize to upper case to avoid duplicating data
        # in IdefixVtkFieldInfo.known_other_fields
        varname = varname.upper()

        if datatype == "SCALARS":
            next(fh)
            retv[varname] = fh.tell()
            read_single_field(fh, shape=shape, skip_data=True)
        elif datatype == "VECTORS":
            for axis in "XYZ":
                vname = f"{varname}_{axis}"
                retv[vname] = fh.tell()
                read_single_field(fh, shape=shape, skip_data=True)
        else:
            raise RuntimeError(f"Unknown datatype {datatype!r}")
        fh.readline()
    return retv
